/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 *
 *     This file is part of Tav4SB.
 *
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 *
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 *
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 *
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine;

import com.wolfram.jlink.Expr;
import com.wolfram.jlink.KernelLink;
import com.wolfram.jlink.MathLinkException;
import com.wolfram.jlink.MathLinkFactory;
import java.io.File;
import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Logger;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.MathematicaProperties;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.Utils;

/**
 *
 * @author Pawel Banasik
 */
public class MEngine {

    private static final Logger logger = Logger.getLogger(MEngine.class);

    private static final MathematicaProperties props = new MathematicaProperties();
    static final String MLINK_IMGFMT = props.getMLinkImageFormat();
    static final String[] MLINK_PARAMS = props.getMLinkParams();
    static final String MSCRIPT_ROOT = props.getScriptRoot();
    static final String MSCRIPT_NOCACHE = props.getScriptNoCache();
    static int engineCount = 0;
    private final KernelLink ml;
    private Set<String> loadedScriptFiles = new HashSet<String>();
    private int id;
    private static final Object lock = new Object();

    static {
        System.setProperty("com.wolfram.jlink.libdir", props.getMLinkLibDir());
    }

    public MEngine() throws MathLinkException {
        try {
            logger.debug("Creating kernel link with params: " + Arrays.toString(MLINK_PARAMS));
            ml = MathLinkFactory.createKernelLink(MLINK_PARAMS);
            logger.debug("Connecting to MathLink: " + ml.toString());
            ml.connect();
            ml.discardAnswer();
        } catch (MathLinkException e) {
            logger.error("Couldn't create or connect to kernel link.", e);
            throw e;
        } catch (Error e) {
            logger.error("Couldn't create or connect to kernel link.", e);
            throw e;
        }

        synchronized (lock) {
            setId(engineCount++);
        }
        logger.debug("Kernel link #" + getId() + " created");
    }

    private KernelLink getKernel() {
        return ml;
    }

    private void setId(int i) {
        this.id = i;
    }

    public int getId() {
        return id;
    }

    public void close() {
        logger.debug("close: (" + getId() + ") engine closed");
        getKernel().close();
    }

    public void release() {
        logger.debug("release: (" + getId() + ") engine release");
        MEngineManager.releaseEngine(this);
    }

    private void loadScriptFile(String name)
            throws MathLinkException, MEngineException
    {
        logger.debug("loadScriptFile: (" + getId() + ") name == " + name);
        if ((MSCRIPT_NOCACHE != null) || !loadedScriptFiles.contains(name)) {
            KernelLink ker = getKernel();
            logger.debug("loadScriptFile: (" + getId() + ") loading script file: " + name);
            // If name is the name of a Mathematica context, ending with a "`"
            // context mark character, then Get operator will process this name
            // to find the file to read. Otherwise get the file from the
            // MSCRIPT_ROOT directory.
            String fileName = name.endsWith("`") ? name : (new File(MSCRIPT_ROOT, name)).getAbsolutePath();
            logger.debug("loadScriptFile: (" + getId() + ") fileName = " + fileName);
            ker.evaluate("<<\"" + fileName + "\"");
            ker.waitForAnswer();
            if (!ker.getSymbol().equalsIgnoreCase("null")) {
                throw new MEngineException("Could not load Mathematica file \"" + fileName + "\"");
            }
            loadedScriptFiles.add(name);
        }
    }

    public void loadScriptFiles(String[] scriptFiles)
            throws MathLinkException, MEngineException
    {
        logger.debug("loadScriptFiles: (" + getId() + ")");
        for (String s : scriptFiles) {
            loadScriptFile(s);
        }
    }

    private void doEvaluate(Object expression) throws MathLinkException {
        logger.debug("doEvaluate: (" + getId() + ") evaluating expression: " + expression);
        if (expression instanceof Expr) {
            getKernel().evaluate((Expr) expression);
        } else {
            getKernel().evaluate((String) expression);
        }
    }

    public void evaluate(Object expression) throws MathLinkException {
        logger.debug("evaluate: (" + getId() + ")");
        doEvaluate(expression);
        getKernel().waitForAnswer();
    }

    public void evaluateAndDiscard(Object expression) throws MathLinkException {
        logger.debug("evaluateAndDiscard: (" + getId() + ")");
        doEvaluate(expression);
        getKernel().discardAnswer();
    }

    public Object getObject() throws MathLinkException {
        logger.debug("getObject: (" + getId() + ")");
        Object result = getKernel().getObject();
        logger.debug("getObject: (" + getId() + ") result = " + result);
        return result;
    }

    public double[] getDoubleArray1() throws MathLinkException {
        logger.debug("getDoubleArray1: (" + getId() + ")");
        double[] result = getKernel().getDoubleArray1();
        StringBuilder sb = new StringBuilder("[");
        for (double d : result) {
            sb.append("" + d + ";");
        }
        sb.append("]");
        logger.debug("getDoubleArray1: (" + getId() + ") result: " + sb.toString());
        return result;
    }

    public double getDouble() throws MathLinkException {
        logger.debug("getDouble: (" + getId() + ")");
        double result = getKernel().getDouble();
        logger.debug("getDouble: (" + getId() + ") result = " + result);
        return result;
    }

    public double[][] getDoubleDoubleArray() throws MathLinkException {
        logger.debug("getDoubleArray2: (" + getId() + ")");
        return getKernel().getDoubleArray2();
    }

    public String evaluateToOutputForm(String s, int i) {
        logger.debug("evaluateToOutputForm: (" + getId() + ")");
        return getKernel().evaluateToOutputForm(s, i);
    }

    public byte[] evaluateToTypeset(String s, int i, boolean b) {
        logger.debug("evaluateToTypeset: (" + getId() + ")");
        return getKernel().evaluateToTypeset(s, i, b);
    }

    public Expr getExpr() throws MathLinkException {
        logger.debug("getExpr: (" + getId() + ")");
        return getKernel().getExpr();
    }

    public byte[] plot(String expression) throws MathLinkException {
        logger.debug("plot(String): (" + getId() + ")");
        getKernel().evaluateToOutputForm("$DefaultImageFormat = \"" + MLINK_IMGFMT + "\"", 0);
        byte[] pngData = doPlot(expression);
        //ml.evaluateToOutputForm("CloseFrontEnd[]", 0);
        return pngData;
    }

    public byte[] plot(Expr expression) throws MathLinkException {
        logger.debug("plot(Expr): (" + getId() + ")");
        getKernel().evaluateToOutputForm("$DefaultImageFormat = \"" + MLINK_IMGFMT + "\"", 0);
        byte[] pngData = doPlot(expression);
        //ml.evaluateToOutputForm("CloseFrontEnd[]", 0);
        return pngData;
    }

    public byte[][] plotMultiple(String[] expression) throws MathLinkException {
        logger.debug("plotMultiple: (" + getId() + ")");
        getKernel().evaluateToOutputForm("$DefaultImageFormat = \"" + MLINK_IMGFMT + "\"", 0);
        byte[][] pngData = new byte[expression.length][];
        for (int i = 0; i < expression.length; i++) {
            pngData[i] = doPlot(expression[i]);
        }
        //ml.evaluateToOutputForm("CloseFrontEnd[]", 0);
        return pngData;
    }

    private byte[] doPlot(String expression) {
        logger.debug("doPlot(String): (" + getId() + ") expresion: " + expression);
        return getKernel().evaluateToTypeset(expression, 0, false);
    }

    private byte[] doPlot(Expr expression) {
        logger.debug("doPlot(Expr): (" + getId() + ") expresion: " + expression);
        return getKernel().evaluateToTypeset(expression, 0, false);
    }

    public void evaluateFunction(String functionName, Object... param) throws MathLinkException {
        logger.debug("evaluateFunction: (" + getId() + ") function: " + functionName);
        for (int i = 0; i < param.length; i++) {
            if (param[i] instanceof BigDecimal[]) {
                param[i] = Utils.bigDecimalArray2DoubleArray((BigDecimal[]) param[i]);
            }
        }

        Expr e = MEngineUtils.fun(functionName, MEngineUtils.objectArray2ExprArray(param));
        evaluate(e);
    }

    public void evaluateFunctionAsText(String functionName, Object... param) throws MathLinkException {
        logger.debug("evaluateFunctionAsText (" + getId() + ") function: " + functionName);
        String command = functionName + "[" + Utils.objectArray2String(param, ',', true) + "]";
        evaluate(command);
    }

    public void discardAnswer() throws MathLinkException {
        logger.debug("discardAnswer (" + getId() + ")");
        getKernel().discardAnswer();
    }
}
