/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc;

import com.wolfram.jlink.Expr;
import com.wolfram.jlink.ExprFormatException;
import com.wolfram.jlink.MathLinkException;
import java.math.BigDecimal;
import java.util.AbstractMap.SimpleEntry;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngine;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineException;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineManager;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineUtils;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.plot.Plot;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.XsdUtils;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NewArrayOfParam;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NewParam;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.ParamRanking;

/**
 *
 * @author trybik
 */
public class CalcSensitivity {

    static final String[] sourceFiles = { "CalcSensitivity.m" };
    private static final Logger logger = Logger
            .getLogger(CalcSensitivity.class);
    private static final String calcSensitivity = "calcSensitivity";

    public static List<ParamRanking> calcSensitivity(
            NewArrayOfParam params,
            List<NewArrayOfParam> acceptableSamples,
            List<NewArrayOfParam> unacceptableSamples,
            boolean generatePlots,
            List<Plot> plotsContainer
    ) throws MathLinkException, MEngineException, ExprFormatException {
        logger.debug(calcSensitivity);
        /* Check and prepare args */

        Map<String, SimpleEntry<String, BigDecimal>> paramsMap = XsdUtils
                .newArrayOfParam2Map(params);
        String[] names = paramsMap.keySet().toArray(new String[0]);
        double[][] acceptableMatrix = XsdUtils.paramArrayList2doubledoubleArray(acceptableSamples, names);
        logger.debug(calcSensitivity + ": Parsed " + acceptableMatrix.length
                + " acceptable samples");
        double[][] unacceptableMatrix = XsdUtils.paramArrayList2doubledoubleArray(unacceptableSamples, names);
        logger.debug(calcSensitivity + ": Parsed " + unacceptableMatrix.length
                + " unacceptable samples");

        /* Acquire engine */
        MEngine engine = null;
        try {
            engine = MEngineManager.getEngine();
            engine.loadScriptFiles(sourceFiles);

            /* Execute function */
            engine.evaluateFunction("CalcSensitivity", acceptableMatrix,
                    unacceptableMatrix);

// //////////////////////////////////////////////////////////////////////////////
// TODO: Banasik mess; refactor
// //////////////////////////////////////////////////////////////////////////////
            /* Get result */
            Expr expr = engine.getExpr();
            logger.debug(calcSensitivity + ": result = " + expr);
            Expr[] result = expr.args();

            List<ParamRanking> res = new LinkedList<ParamRanking>();

            Expr[] functionNames = result[0].args();
            logger.debug(calcSensitivity + ": created " + functionNames.length
                    + " parameter sensitivity rankings with functions "
                    + Arrays.toString(functionNames));

            for (Expr func : functionNames) {
                ParamRanking pr = new ParamRanking();
                pr.setSortedParamList(new NewArrayOfParam());
                pr.setName(func.asString());
                res.add(pr);
                logger.debug(calcSensitivity + ": ranking #"
                        + ": sensitivity function is " + pr.getName());
            }

            Expr[] rankingsExpr = result[1].args();
            logger.debug(calcSensitivity + ": " + rankingsExpr.length
                    + " rankings expressions are "
                    + Arrays.toString(rankingsExpr));

            for (int i = 0; i < names.length; i++) {
                Expr[] paramResult = rankingsExpr[i].args();
                Expr[] sensitivities = paramResult[0].args();
                String paramName = names[i];
                String paramType = paramsMap.get(paramName).getKey();
                for (int j = 0; j < res.size(); j++) {
                    res.get(j).getSortedParamList().getParam().add(
                        XsdUtils.createNewParam(paramName, paramType,
                                new BigDecimal(sensitivities[j].asDouble())
                        )
                    );
                }

                if (generatePlots) {

                    String yLabel = paramName + " (" + paramType + ")";

//                    byte[] frequencyPlot = engine.plot(MEngineUtils.fun(
//                            "PlotFrequency", new Expr[] {
//                                    paramResult[1],
//                                    paramResult[2],
//                                    new Expr(yLabel),
//                                    new Expr("Frequency") }));
//                    Plot freqPlot = new Plot(2);
//                    freqPlot.setData(frequencyPlot);
//                    plotsContainer.add(freqPlot);

                    byte[] cumulativeFrequencyPlot = engine.plot(MEngineUtils.fun(
                            "PlotFrequency", new Expr[] {
                                    paramResult[1],
                                    paramResult[3],
                                    new Expr(yLabel),
                                    new Expr("Cumulative frequency") }));
                    Plot cumulativPlot = new Plot(2);
                    cumulativPlot.setData(cumulativeFrequencyPlot);
                    plotsContainer.add(cumulativPlot);
                }
            }

            /* Build response */
            for (ParamRanking pr : res) {
                Collections.sort(
                    pr.getSortedParamList().getParam(),
                    new Comparator<NewParam>() {
                        public int compare(NewParam param1, NewParam param2) {
                            return 1-param1.getValue().compareTo(param2.getValue());
                        }
                    });
            }

            logger.debug(calcSensitivity);
            return res;
////////////////////////////////////////////////////////////////////////////////
        } catch (MathLinkException mle) {
            logger.error(calcSensitivity, mle);
            throw mle;
        } catch (ExprFormatException efe) {
            logger.error(calcSensitivity, efe);
            throw efe;
        } finally {
            if (engine != null) {
                logger.debug(calcSensitivity + ": relasing engine");
                engine.release();
            }
        }
    }
}
