/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
// TODO: mv to pl.edu.mimuw.bioputer.tav4sb.logic.java.proc
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc;

import com.wolfram.jlink.Expr;
import com.wolfram.jlink.MathLinkException;
import java.util.LinkedList;
import java.util.List;
import org.apache.log4j.Logger;

import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngine;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineManager;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineUtils;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.Utils;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.XsdUtils;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Timepoint;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Trajectories;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Variable;

/**
 *
 * @author michal
 */
public class SelectObservables {

    public static final String IDENTITY = "IDENTITY";
    private static final Logger logger = Logger.getLogger(SelectObservables.class);
    private static final String selectObservables = "selectObservables";
    private static final String checkParams = "checkParams";

    static public Trajectories selectObservables(String method, Trajectories trajectories, List<String> names, List<String> formulas) throws MathLinkException {
        logger.debug(selectObservables);

        checkParams(method, trajectories, names, formulas);

        if (IDENTITY.equalsIgnoreCase(method)) {
            logger.debug(selectObservables + ": Acting like identity function");

            return trajectories;
        }

        List<Variable> observables = new LinkedList<Variable>();
        for (int i = 0; i < formulas.size(); i++) {
            Variable v = new Variable();
            v.setName((names == null) ? ("observable" + i) : names.get(i));
            v.setFormula(formulas.get(i).trim());
            observables.add(v);
        }

        logger.debug(selectObservables + ": Created " + formulas.size() + " observables");

        List<Variable> species = trajectories.getVariableList().getVariable();
        IndexedVariable[] speciesIdx = new IndexedVariable[species.size()];
        for (int i = 0; i < species.size(); i++) {
            speciesIdx[i] = new IndexedVariable("$" + species.get(i).getName(), species.get(i), i + 1);
        }

        for (int i = 0; i < species.size(); i++) {
            int min = i;
            for (int j = i + 1; j < species.size(); j++) {
                if (speciesIdx[j].name.startsWith(speciesIdx[min].name)) {
                    min = j;
                }
            }

            if (min != i) {
                IndexedVariable tmp = speciesIdx[i];
                speciesIdx[i] = speciesIdx[min];
                speciesIdx[min] = tmp;
            }
        }

        for (int j = 0; j < species.size(); j++) {
            speciesIdx[j].name = "\\" + speciesIdx[j].name;
            for (int i = 0; i < formulas.size(); i++) {
                String s = " abc ";
                formulas.set(i, formulas.get(i).replaceAll(speciesIdx[j].name, "#" + speciesIdx[j].index));
            }
        }

        StringBuilder transformFunction = new StringBuilder("{");
        transformFunction.append(Utils.objectList2String(formulas, ',', false));
        transformFunction.append("}&");

        logger.debug(selectObservables + ": Generated transform function: " + transformFunction);

        Timepoint[] timepoints = trajectories.getTimepointList().getTimepoint().toArray(new Timepoint[0]);
        double[][] matrix = XsdUtils.timepoints2doubledoubleArray(timepoints);

        MEngine engine = MEngineManager.getEngine();
        engine.evaluateAndDiscard("f = " + transformFunction.toString() + "; 0");
        engine.evaluate(MEngineUtils.fun("Apply",
                new Expr[]{
                    MEngineUtils.symbol("f"),
                    new Expr(matrix),
                    new Expr(new double[]{1})
                }));

        double[][] resultMatrix = engine.getDoubleDoubleArray();

        engine.release();

        String[] resultValues = Utils.doubledoubleArray2StringArray(resultMatrix, ' ');

        for (int i = 0; i < timepoints.length; i++) {
            timepoints[i].setValues(resultValues[i]);
        }

        trajectories.getVariableList().getVariable().clear();
        trajectories.getVariableList().getVariable().addAll(observables);

        logger.debug(selectObservables);

        return trajectories;
    }

    static class IndexedVariable {

        public String name;
        public Variable Variable;
        public int index;

        public IndexedVariable(String name, Variable Variable, int index) {
            this.name = name;
            this.Variable = Variable;
            this.index = index;
        }
    }

    private static void raiseException(String message) {
        logger.debug(checkParams + ": " + message);
        throw new IllegalArgumentException(message);
    }

    private static void checkParams(String method, Trajectories trajectories, List<String> names, List<String> formulas) {
        if (method == null) {
            raiseException("method is null");
        }
        if (trajectories == null) {
            raiseException("trajectories is null");
        }
        if (names == null) {
            raiseException("names is null");
        }
        if (formulas == null) {
            raiseException("formulas is null");
        }
        if (trajectories.getTimepointList().getTimepoint().isEmpty()) {
            raiseException("empty timepoints list");
        }
        if (trajectories.getVariableList().getVariable().isEmpty()) {
            raiseException("empty variables list");
        }
        if (!IDENTITY.equalsIgnoreCase(method)) {
            if (formulas.size() > names.size()) {
                raiseException("more formulas than names");
            }
        }
    }
}
