/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 *
 *     This file is part of Tav4SB.
 *
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 *
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 *
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 *
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils;

import java.math.BigDecimal;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import edu.emory.mathcs.backport.java.util.Arrays;

import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NewArrayOfParam;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NewParam;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NumericArray;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Param;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Timepoint;

/**
 *
 * @author trybik
 *
 */
public class XsdUtils {

    public static interface ObjectExtractor {

        Object value(int n);

        int nvalues();
    }

    public static BigDecimal[] extractParamLowerBounds(final List<Param> params) {
        return XsdUtils.ExtractBigDecimalArray(new XsdUtils.ObjectExtractor() {

            public BigDecimal value(int n) {
                return params.get(n).getLowerBound();
            }

            public int nvalues() {
                return params.size();
            }
        });
    }

    public static BigDecimal[] extractParamUpperBounds(final List<Param> params) {
        return XsdUtils.ExtractBigDecimalArray(new XsdUtils.ObjectExtractor() {

            public BigDecimal value(int n) {
                return params.get(n).getUpperBound();
            }

            public int nvalues() {
                return params.size();
            }
        });
    }

    public static BigDecimal[] extractParamValues(final List<Param> params) {
        return XsdUtils.ExtractBigDecimalArray(new XsdUtils.ObjectExtractor() {

            public BigDecimal value(int n) {
                return params.get(n).getValue();
            }

            public int nvalues() {
                return params.size();
            }
        });
    }



    public static interface MatrixRowStringExtractor {

        String row(int n);

        int nrows();
    }

    public static double[][] StringMatrix2doubledoubleArray(XsdUtils.MatrixRowStringExtractor e) {
        double[][] result = new double[e.nrows()][];
        for (int i = 0; i < e.nrows(); i++) {
            result[i] = XsdUtils.doubleString2doubleArray(e.row(i));
        }
        return result;
    }

    public static double[] doubleString2doubleArray(String s) {
        String[] doubles = s.split(" ");
        double[] result = new double[doubles.length];
        for (int i = 0; i < doubles.length; i++) {
            result[i] = Double.parseDouble(doubles[i]);
        }
        return result;
    }

    public static double[][] timepoints2doubledoubleArray(final Timepoint[] t) {
        return XsdUtils.StringMatrix2doubledoubleArray(new XsdUtils.MatrixRowStringExtractor() {

            public String row(int n) {
                return t[n].getValues();
            }

            public int nrows() {
                return t.length;
            }
        });
    }

    public static NumericArray doubleArray2NumericArray(double[] array) {
        NumericArray na = new NumericArray();
        for (int i = 0; i < array.length; i++) {
            na.getValue().add(new BigDecimal(array[i]));
        }
        return na;
    }



    public static double[][] paramArrayList2doubledoubleArray(
            List<NewArrayOfParam> paramArrayList, String[] names) {
        List<double[]> resultList = new ArrayList<double[]>(paramArrayList.size());
        double[] row;
        for (NewArrayOfParam paramArray: paramArrayList) {
            row = newArrayOfParam2doubleArray(paramArray, names);
            if (row != null) resultList.add(row);
        }
        double[][] result = new double[resultList.size()][];
        for (int i = 0; i < resultList.size(); i++)
            result[i] = resultList.get(i);
        return result;
    }

    public static double[] newArrayOfParam2doubleArray(
            NewArrayOfParam paramArray, String[] names) {
        Map<String, SimpleEntry<String, BigDecimal>> paramMap = newArrayOfParam2Map(paramArray);

        if(!paramMap.keySet().containsAll(Arrays.asList(names)))
            return null;

        double[] ret = new double[names.length];
        for (int i = 0; i < ret.length; i++) {
            ret[i] = paramMap.get(names[i]).getValue().doubleValue();
        }
        return ret;
    }

    public static Map<String, SimpleEntry<String, BigDecimal>> newArrayOfParam2Map(
            NewArrayOfParam paramArray) {
        List<NewParam> params = paramArray.getParam();
        // preserve order of params
        Map<String, SimpleEntry<String, BigDecimal>> ret = new LinkedHashMap<String, SimpleEntry<String, BigDecimal>>(params.size());
        for (NewParam param : params) {
            if (!ret.containsKey(param.getName())) //ignore duplicates
                ret.put(param.getName(), new SimpleEntry<String, BigDecimal>(param.getType(), param.getValue()));
        }
        return ret;
    }

    //temporary
    public static NewParam createNewParam(String n, String t, BigDecimal v) {
        NewParam ret = new NewParam();
        ret.setName(n);
        ret.setType(t);
        ret.setValue(v);
        return ret;
    }

    public static NewParam param2NewParam(Param p) {
        String name;
        String type;
        if (p.getReaction() == null) {
            type = "species";
            name = p.getName();
        } else {
            type = "kinetic";
            name = p.getReaction().length() > 0 ? p.getReaction() + "." + p.getName(): p.getName();
        }
        return createNewParam(name, type, p.getValue());
    }



    public static double[] newArrayOfParam2doubleArray(
            NewArrayOfParam paramArray) {
        return null;
    }



    public static BigDecimal[] ExtractBigDecimalArray(XsdUtils.ObjectExtractor e) {
        BigDecimal[] result = new BigDecimal[e.nvalues()];
        for (int i = 0; i < e.nvalues(); i++) {
            result[i] = (BigDecimal) e.value(i);
        }
        return result;
    }

    public static String[] ExtractStringArray(XsdUtils.ObjectExtractor e) {
        String[] result = new String[e.nvalues()];
        for (int i = 0; i < e.nvalues(); i++) {
            result[i] = (String) e.value(i);
        }
        return result;
    }

}
