/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 *
 *     This file is part of Tav4SB.
 *
 *     Tav4SB is free software: you can redistribute it and/or modify it under
 *     the terms of the GNU Affero General Public License as published by the
 *     Free Software Foundation, either version 3 of the License, or (at your
 *     option) any later version.
 *
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 *
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 *
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.worker;

import com.wolfram.jlink.MathLinkException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import pl.edu.mimuw.bioputer.tav4sb.consumer.task.AbstractWorker;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.engine.MEngineException;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.plot.AbstractPlot;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.plot.Plot;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.proc.plot.PlotPoint;
import pl.edu.mimuw.bioputer.tav4sb.logic.mathematica.utils.XsdMathTypeUtils;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.ArrayOfPoint;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.MathType;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Option;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.PlotAxis;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.Point;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.MathPlotParams;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.MathPlotResult;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.NumericValue;
import pl.edu.mimuw.bioputer.tav4sb.utils.xml.XmlTranslator;

/**
 *
 * @author michal
 */
public class MathPlotWorker extends
        AbstractWorker<MathPlotParams, MathPlotResult> {

    public MathPlotWorker(
            XmlTranslator<MathPlotResult, MathPlotParams> translator) {
        super(translator);
    }

    /**
     * Transform Point to coordinate name -> value map.
     */
    private Map<String, BigDecimal> pointToMap(Point p) {
        Map<String, BigDecimal> result = new HashMap<String, BigDecimal>();
        for (NumericValue c : p.getCoordinate()) {
            result.put(c.getName(), c.getValue());
        }
        return result;
    }

    /**
     * Filter point coordinates according to axes names. If point is missing
     * coordinate than return null.
     */
    private List<BigDecimal> extractPointValues(Point p, List<PlotAxis> axes) {
        Map<String, BigDecimal> point = pointToMap(p);
        List<BigDecimal> result = new ArrayList<BigDecimal>(axes.size());
        for (PlotAxis axis : axes) {
            String coordinateName = axis.getValue();
            if (!point.containsKey(coordinateName)) {
                result = null;
                break;
            }
            result.add(point.get(coordinateName));
        }
        return result;
    }

    /**
     * Return plot representing data series line (2D) or surface (3D).
     */
    private AbstractPlot getPlot(ArrayOfPoint data, List<PlotAxis> axes) {
        List<Point> points = data.getPoint();
        AbstractPlot result = AbstractPlot.createPlot(axes.size());
        for (Point point : points) {
            List<BigDecimal> pValues = extractPointValues(point, axes);
            if (pValues != null)
                result.addPoint(PlotPoint.getPlotPoint(pValues));
        }
        return result;
    }

    private List<AbstractPlot> getPlots(List<ArrayOfPoint> data,
            List<PlotAxis> axes) {
        List<AbstractPlot> result = new ArrayList<AbstractPlot>(data.size());
        for (ArrayOfPoint series : data) {
            result.add(getPlot(series, axes));
        }
        return result;
    }

    /**
     * Return axis label; if no label, then take the value.
     */
    private String getAxisLabel(PlotAxis axis) {
        if (axis.getLabel() != null) {
            return axis.getLabel();
        }
        return axis.getValue();
    }

    private List<String> getAxesLabels(List<PlotAxis> axes) {
        List<String> labels = new ArrayList<String>(3);
        for (PlotAxis axis : axes) {
            labels.add(getAxisLabel(axis));
        }
        return labels;
    }

    /**
     * Create "AxesLabel" option from axes labels (axis/@label) or values.
     * @param data
     * @param options
     */
    private void apendAxesLabelOpt(List<PlotAxis> axes, List<Option> options) {
        // TODO Auto-generated method stub
        pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.List axesLabel =
            new pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.List();
        for (String label: getAxesLabels(axes)) {
            axesLabel.getElement().add(XsdMathTypeUtils.fromString(label));
        }
        Option axesLabelOpt = new Option();
        axesLabelOpt.setName("AxesLabel");
        axesLabelOpt.setValue(axesLabel);
        options.add(axesLabelOpt);
    }

    /**
     * Create "PlotLegend" option from data series names (data/@name). If
     * created append it to the passed options list.
     */
    private boolean apendPlotLegendOpt(List<ArrayOfPoint> data, List<Option> options) {
        boolean apendPlotLegendOpt = false;
        pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.List plotLegend =
            new pl.edu.mimuw.bioputer.tav4sb.services.xsd.mathematica.List();
        for (ArrayOfPoint dataSeries: data) {
                String name = dataSeries.getName();
                if (name != null) {
                    apendPlotLegendOpt = true;
                } else {
                    name = "";
                }
                plotLegend.getElement().add(XsdMathTypeUtils.fromString(name));
        }
        if (apendPlotLegendOpt) {
            Option plotLegendOpt = new Option();
            plotLegendOpt.setName("PlotLegend");
            plotLegendOpt.setValue(plotLegend);
            options.add(plotLegendOpt);
        }
        return apendPlotLegendOpt;
    }

    private Map<String, MathType> getOptionsMap(List<PlotAxis> axes,
            List<ArrayOfPoint> data, List<Option> options) {

        apendAxesLabelOpt(axes, options);
        apendPlotLegendOpt(data, options);

        Map<String, MathType> result = new HashMap<String, MathType>(options
                .size());

        for (Option option : options) {
            // ignore repeating options, in particular explicit "AxesLabel" and
            // "PlotLegend" opts declarations have precedence
            if (!result.containsKey(option.getName())) {
                result.put(option.getName(), option.getValue());
            }
        }
        return result;
    }

    @Override
    public MathPlotResult doExecute(MathPlotParams param)
            throws MathLinkException, MEngineException {

        List<PlotAxis> axes = param.getAxis();
        List<ArrayOfPoint> data = param.getData();
        Plot p = new Plot(axes.size(), getOptionsMap(axes, data, param.getOption()));
        p.addPlots(getPlots(data, axes));

        p.draw();

        MathPlotResult result = new MathPlotResult();
        result.getPlot().add(p.getData());

        return result;
    }
}
