/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.odesolver;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import edu.emory.mathcs.backport.java.util.Arrays;
import pl.edu.mimuw.bioputer.tav4sb.logic.odesolver.utils.OdeSolverProperties;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.ArrayOfTimepointValues;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.ArrayOfVariable;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.TimepointValues;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.Trajectories;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.FiniteValue;
import pl.edu.mimuw.bioputer.tav4sb.services.xsd.odesolver.Variable;
import pl.edu.mimuw.bioputer.tav4sb.utils.cli.SysCmd;

/**
 *
 * @author michal
 */
public class OdeSolver {

    private static final Logger logger = Logger.getLogger(OdeSolver.class);

    private static final OdeSolverProperties props = new OdeSolverProperties();
    static final BigDecimal DEFAULT_ATOL = props.getATOL();
    static final BigDecimal DEFAULT_RTOL = props.getRTOL();
    static final String PARSER_VARLISTHEADER = props.getParserVarListHeader();
    static final String PARSER_VARLISTSEPARATOR = props.getParserVarListSeparator();
    static final String PARSER_VALLISTSEPARATOR = props.getParserValListSeparator();
    static final String PARSER_CONCENTRATIONSSEPARATOR = props.getParserConcentrationSeparator();
    static final String PARSER_SENSITIVITIESSEPARATOR = props.getParserSensitivitiesSeparator();

    private final String command;

    public OdeSolver(String COMMAND) {
        this.command = COMMAND;
    }

    public Trajectories run(String input, BigDecimal simt, BigDecimal simtp, BigDecimal atol, BigDecimal rtol, Boolean sensitivityAnalysis) throws IOException, InterruptedException {
        logger.debug("begin");

        if (simtp == null) {
            logger.debug("simtp not set, using default value = simt");
            simtp = simt;
        }
        if (atol == null) {
            logger.debug("atol not set, using default value " + DEFAULT_ATOL);
            atol = DEFAULT_ATOL;
        }
        if (rtol == null) {
            logger.debug("rtol not set, using default value " + DEFAULT_RTOL);
            rtol = DEFAULT_RTOL;
        }

        checkParams(input, simt, simtp, atol, rtol, sensitivityAnalysis);

        String dataSeparator = Boolean.TRUE.equals(sensitivityAnalysis) ? PARSER_SENSITIVITIESSEPARATOR : PARSER_CONCENTRATIONSSEPARATOR;

        File tmpFile = File.createTempFile("tav4sb", ".xml");
        Map<String, String> cmdOutMap;
        BufferedReader stdOutByLineReader = null;

        try {
            FileOutputStream writer = new FileOutputStream(tmpFile);
            writer.write(input.getBytes());
            writer.close();

            String[] c = new String[]{

                getCommand(),tmpFile.getName(),
                "--mpath",
                tmpFile.getParent(),
                "--time",
                simt.toString(),
                "--printstep",
                simtp.toString(),
                "--error",
                atol.toString(),
                "--rerror",
                rtol.toString(),
                Boolean.TRUE.equals(sensitivityAnalysis) ? "--sensitivity" : ""
            };

            cmdOutMap = SysCmd.execute(c);
            stdOutByLineReader = new BufferedReader(new StringReader(cmdOutMap.get(SysCmd.OUT_KEY)));
        } finally {
            tmpFile.delete();
        }

        String line = stdOutByLineReader.readLine();
        /*
         * line == null -- OdeSolver error
         */
        if (line == null) {
            logger.debug("No odeSolver output.");
            throw new OdeSolverError(cmdOutMap.get(SysCmd.ERROR_KEY));
        }
        if (!line.startsWith(PARSER_VARLISTHEADER)) {
            logger.error("expected odeSolver output to start with header \"" + PARSER_VARLISTHEADER +"\"");
            throw new ParseErrorException(line, PARSER_VARLISTHEADER);
        }
        line = line.substring(PARSER_VARLISTHEADER.length());
        String[] variableNames = line.split(PARSER_VARLISTSEPARATOR);

        Trajectories result = new Trajectories();
        result.setTimepoints(new ArrayOfTimepointValues());
        result.setVariables(new ArrayOfVariable());

        for (int i = 0; i < variableNames.length; i++) {
            Variable v = new Variable();
            /*
             *  where is formula name setting ?!?!
             *        v.setFormula(String)
             */
            v.setName(variableNames[i]);

            result.getVariables().getVariable().add(v);
        }

        logger.debug("Parsed " + variableNames.length + " variable names");

        if (!(line = stdOutByLineReader.readLine()).equals(dataSeparator)) {
            logger.error("expected separator \"" + dataSeparator + "\" after odeSolver output header");
            throw new ParseErrorException(line, dataSeparator);
        }

        while (!(line = stdOutByLineReader.readLine()).equals(dataSeparator)) {
            if (line.equals("")) {
                continue;
            }

            int spacePos = line.indexOf(' ');
            if (spacePos == -1) {
                logger.error("space within the line \"" + line + "\"");
                throw new ParseErrorException(line, "space within the line");
            }

            TimepointValues tp = new TimepointValues(/*line.substring(0, spacePos), line.substring(spacePos + 1).trim()*/);

            TimepointValues.Time t = new TimepointValues.Time();
            t.setValue(new BigDecimal(line.substring(0, spacePos)));
            t.setName("time");
            tp.setTime(t);

            String[] variablesValues = line.substring(spacePos + 1).trim().split(PARSER_VALLISTSEPARATOR);
            if (variablesValues.length != variableNames.length) {
                logger.info("Skipping line \"" + line + "\"; incompatible number of values for variables " + Arrays.toString(variableNames) + ".");
            } else {
                List<FiniteValue> variableList = tp.getVariable();
                FiniteValue val;
                for (int j = 0; j < variablesValues.length; j++) {
                    val = new FiniteValue();
                    val.setName(variableNames[j]);
                    val.setValue(new BigDecimal(variablesValues[j]));
                    variableList.add(val);
                }
                result.getTimepoints().getTimepoint().add(tp);
            }
        }
        logger.debug("Succesfully parsed " + result.getTimepoints().getTimepoint().size() + " timepoints");
        logger.debug("end");
        return result;
    }

    private void raiseException(String errMsg) {
        logger.error(errMsg);
        throw new IllegalArgumentException(errMsg);
    }

    private void checkParams(String input, BigDecimal simt, BigDecimal simtp, BigDecimal atol, BigDecimal rtol, Boolean sensitivityAnalysis) {
        if (input == null) {
            raiseException("input is null");
        }
        if (simt == null) {
            raiseException("simt is null");
        }
        if (simtp == null) {
            raiseException("simtp is null");
        }
        if (atol == null) {
            raiseException("atol is null");
        }
        if (rtol == null) {
            raiseException("rtol is null");
        }
        if (sensitivityAnalysis == null) {
            raiseException("sensitivityAnalysis is null");
        }
        if (simtp.signum() <= 0) {
            raiseException("simtp <= 0");
        }
    }

    /**
     * @return the command
     */
    public String getCommand() {
        return command;
    }
}
