/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.logic.prism;

import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.apache.log4j.Logger;
import pl.edu.mimuw.bioputer.tav4sb.utils.cli.CmdInOutTranslator;
import pl.edu.mimuw.bioputer.tav4sb.utils.cli.ObjectSysCmd;
import pl.edu.mimuw.bioputer.tav4sb.utils.file.RemovableFilesHolder;
import pl.edu.mimuw.bioputer.tav4sb.utils.file.SaveableStringContent;

/**
 * Wrapper of prism model checker command line tool.
 * Attention: Prism command objects are not thread-safe.
 *
 * @author michal
 */
public class Prism extends ObjectSysCmd<PrismArgs, PrismResult> {

    private static final String version = "-version";

    /**
     * A statefull translator - not thread-safe.
     */
    private static class PrismInOutTranslator extends RemovableFilesHolder implements CmdInOutTranslator<PrismArgs, PrismResult> {

        private static final Logger logger = Logger.getLogger(PrismInOutTranslator.class);
        private static final String constant = "-const";
        private static final String fixdl = "-fixdl";
        private Map<PrismConstant, BigDecimal> constants;

        /**
         * Transform constants map to string proper for command line prism execution.
         *
         * @param Map<PrismConstant, BigDecimal> map of constants and values.
         * @return String representation of a map of constants.
         */
        private String constantsMapToString(Map<PrismConstant, BigDecimal> map) {
            String entry = null;
            StringBuilder sb = new StringBuilder();
            for (Entry<PrismConstant, BigDecimal> e : map.entrySet()) {
                if (entry != null) {
                    sb.append(entry + ",");
                }
                entry = e.getKey().getLabel() + "=" + e.getValue();
                e.getKey();
                e.getValue();
            }
            sb.append(entry);
            return sb.toString();
        }

        private File prepareTempFile(String sufix, SaveableStringContent content) {
            File f;
            try {
                f = File.createTempFile("tav4sb", sufix);
            } catch (IOException ex) {
                logger.error("Couldn't create temporary PRISM file.", ex);
                throw new RuntimeException("Couldn't create temporary PRISM file.", ex);
            }
            try {
                // note: no need to check if f is null
                getFiles().add(f);
                content.save(f);
            } catch (IOException ex) {
                logger.error("Couldn't write to temporary PRISM file" + f.getAbsolutePath() + ".", ex);
                throw new RuntimeException("Couldn't write to temporary PRISM file" + f.getAbsolutePath() + ".", ex);
            }
            logger.debug("Created and saved temporary file " + f.getAbsolutePath() + ".");
            return f;
        }

        public String[] toArgs(PrismArgs args) {
            logger.debug("(Pre) Current # of temp files holded = " + getFiles().size());
            File modelFile = prepareTempFile("model.sm", args.getModel());
            File propertiesFile = prepareTempFile("properties.csl", args.getProperties());
            logger.debug("(Post) Current # of temp files holded = " + getFiles().size());
            constants = args.getConstValue();
            List<String> list = new LinkedList<String>();
            list.add(modelFile.getAbsolutePath());
            list.add(propertiesFile.getAbsolutePath());
            list.add(fixdl);
            if (!constants.isEmpty()) {
                list.add(constant);
                list.add(constantsMapToString(constants));
            }
            return list.toArray(new String[0]);
        }

        public PrismResult fromStdOut(String out) {
            return new PrismResult(out, constants);
        }
    };

    public Prism(String cmd) {
        // using non thread-safe translator, thus command obj is not thread-safe
        super(cmd, new PrismInOutTranslator());
    }

    /**
     * Version of the PRISM model checker.
     * @return String with the PRISM version
     * @throws IOException
     * @throws InterruptedException
     */
    public String getVersion() throws IOException, InterruptedException {
        String[] args = {version};
        return run(args);
    }

    /**
     * Simulate prism model and evaluate properties. Set model and properties
     * constants to values from given map.
     * @param model prism model
     * @param properties prism csl properties
     * @param constValue map of constants values
     * @return prism simulation result
     * @throws java.io.IOException
     * @throws InterruptedException
     */
    public PrismResult run(PrismModel model, PrismProperties properties, Map<PrismConstant, BigDecimal> constValue) throws IOException, InterruptedException {
        return run(new PrismArgs(model, properties, constValue));
    }
}
