/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.producer;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.jms.Destination;
import javax.jms.MapMessage;
import org.apache.activemq.ActiveMQConnectionFactory;
import org.apache.log4j.Logger;

public class Producer {

    public static final String ANONYMOUS_TASK_ID = "";
    private static final Logger logger = Logger.getLogger(Producer.class);
    private String queueName;
    private String brokerUrl;
    private ConnectionFactory factory;
    private Connection connection;
    private Session session;
    private MessageProducer producer;
    private MessageConsumer consumer;
    private Destination destination;
    private Destination replyTo;

    private void initialize(String brokerURL, String queueName) throws JMSException {
        this.queueName = queueName;
        this.brokerUrl = brokerURL;
        this.factory = new ActiveMQConnectionFactory(getBrokerUrl());
        try {
            this.connection = getFactory().createConnection();
            getConnection().start();
            this.session = getConnection().createSession(false, Session.AUTO_ACKNOWLEDGE);
            this.destination = getSession().createQueue(getQueueName());
            this.producer = getSession().createProducer(getDestination());
        } catch (JMSException ex) {
            logger.error("JMSException", ex);
            throw ex;
        }
    }

    public Producer(String brokerURL, String queueName) throws JMSException {
        initialize(brokerURL, queueName);
        this.replyTo = getSession().createTemporaryQueue();
        this.consumer = getSession().createConsumer(getReplyTo());
    }

    public Producer(String brokerURL, String queueName, String replyQueueName) throws JMSException {
        initialize(brokerURL, queueName);
        this.replyTo = getSession().createQueue(replyQueueName);
        this.consumer = getSession().createConsumer(getReplyTo());
    }

    public String getQueueName() {
        return this.queueName;
    }

    public String getBrokerUrl() {
        return this.brokerUrl;
    }

    public ConnectionFactory getFactory() {
        return this.factory;
    }

    public Connection getConnection() {
        return this.connection;
    }

    public MessageProducer getProducer() {
        return this.producer;
    }

    public MessageConsumer getConsumer() {
        return this.consumer;
    }

    public Session getSession() {
        return this.session;
    }

    public Destination getDestination() {
        return this.destination;
    }

    public Destination getReplyTo() {
        return this.replyTo;
    }

    public void send(Map<String, String> map) throws JMSException {
        MapMessage message = getSession().createMapMessage();
        message.setJMSReplyTo(getReplyTo());
        logger.info("Sending message");
        for (Entry<String, String> entry : map.entrySet()) {
            logger.info("key: " + entry.getKey());
            logger.info("value: " + entry.getValue());
            message.setString(entry.getKey(), entry.getValue());
        }
        getProducer().send(message);
    }

    public void send(String type, String text, String id) throws JMSException {
        Map<String, String> map = new HashMap<String, String>();
        map.put("ID", id);
        map.put("TYPE", type);
        map.put("TEXT", text);
        send(map);
    }

    public void send(String type, String text) throws JMSException {
        send(type, text, ANONYMOUS_TASK_ID);
    }

    protected String doReceive(Message m) throws JMSException {
        String errMsg;
        if (m instanceof MapMessage) {
            MapMessage map = (MapMessage) m;
            if (map.itemExists("EXCEPTION")) {
                errMsg = "EXCEPTION\n" + map.getString("EXCEPTION");
                logger.error(errMsg);
                throw new IllegalStateException(errMsg);
            }
            if (map.itemExists("RESULT")) {
                logger.info("RESULT");
                logger.info(map.getString("RESULT"));
                return map.getString("RESULT");
            } else {
                errMsg = "no RESULT property";
                logger.error(errMsg);
                throw new IllegalStateException(errMsg);
            }
        } else {
            errMsg = "MapMessage expected";
            logger.error(errMsg);
            throw new IllegalStateException();
        }
    }

    public String receive() throws JMSException {
        return doReceive(getConsumer().receive());
    }

    public String receiveNoWait() throws JMSException {
        return doReceive(getConsumer().receiveNoWait());
    }

    public String receive(long timeout) throws JMSException {
        return doReceive(getConsumer().receive(timeout));
    }

    public void close() throws JMSException {
        logger.debug("closing producer");
        if (getConnection() != null) {
            getConnection().close();
        }
    }
}
