/*******************************************************************************
 * Copyright (C) 2011 Mikołaj Rybiński
 * 
 *     This file is part of Tav4SB.
 * 
 *     Tav4SB is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Affero General Public License as published by the Free
 *     Software Foundation, either version 3 of the License, or (at your option)
 *     any later version.
 * 
 *     Tav4SB is distributed in the hope that it will be useful, but WITHOUT ANY
 *     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *     FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
 *     more details.
 * 
 *     You should have received a copy of the GNU Affero General Public License
 *     along with Tav4SB. If not, see <http://www.gnu.org/licenses/>.
 * 
 *     Contributors:
 *         Paweł Banasik - initial implementation
 *         Michał Lula - grid implementation
 *         Mikołaj Rybiński - implementation and maintenance
 ******************************************************************************/
package pl.edu.mimuw.bioputer.tav4sb.register;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import javax.jms.JMSException;
import javax.jms.MapMessage;
import javax.jms.Message;
import org.apache.log4j.Logger;

/**
 *
 * @author michal
 */
public class Register {

    private static final Logger logger = Logger.getLogger(Register.class);
    private static final Register instance = new Register();
    private final Object factoryLock = new Object();
    private final Map<String, ComputationFactory> computationFactory = new HashMap<String, ComputationFactory>();
    private final Object activeLock = new Object();
    private final Map<String, Computation> activeComputations = new HashMap<String, Computation>();
    private final Object completedLock = new Object();
    private final List<Computation> completedComputation = new LinkedList<Computation>();
    private final Map<String, List<Message>> unregisteredList = new HashMap<String, List<Message>>();
    private final Object idLock = new Object();
    private int idCounter = 0;

    private Register() {
    }

    public String getId() {
        int id;
        synchronized (idLock) {
            id = idCounter;
            idCounter += 1;
        }
        return String.valueOf(id);
    }

    public static synchronized Register getInstance() {
        return instance;
    }

    private Map<String, ComputationFactory> getComputationFactory() {
        return computationFactory;
    }

    private Map<String, Computation> getActiveComputations() {
        return activeComputations;
    }

    private Map<String, List<Message>> getUnregisteredMessages() {
        return unregisteredList;
    }

    public void addComputationFactory(ComputationFactory factory) {
        logger.debug("new computation factory of type: " + factory.getType());
        synchronized (factoryLock) {
            getComputationFactory().put(factory.getType(), factory);
        }
    }

    public void register(String type, String id, int quantity, String email) throws JMSException {
        logger.info("Registering new task: id="+id+", type="+type+", quantity="+quantity+", to="+email+".");
        ComputationFactory factory = null;
        synchronized (factoryLock) {
            factory = getComputationFactory().get(type);
        }
        if (factory != null) {
            Computation comp = factory.newComputation(id, quantity, email);
            List<Message> unreg = new LinkedList<Message>();
            synchronized (activeLock) {
                getActiveComputations().put(id, comp);
                if (getUnregisteredMessages().containsKey(id)) {
                    unreg.addAll(getUnregisteredMessages().remove(id));
                }
            }
            for (Message answer : unreg) {
                addAnswer(answer);
            }
        } else {
            logger.error("Unsuported task type " + type);
        }
    }

    public void addAnswer(Message message) throws JMSException {
        if (message instanceof MapMessage) {
            MapMessage map = (MapMessage) message;

            Computation c = null;
            String id = map.getString("ID");
            logger.debug("Received answer for task(id=" + id + ")");
            synchronized (activeLock) {
                if (getActiveComputations().containsKey(id)) {
                    c = getActiveComputations().get(id);
                    try {
                        c.addAnswer(map);
                    } catch (Exception e) {
                        logger.error("exception", e);
                    } finally {
                        if (c.isComplete()) {
                            logger.debug("Task(id=" + id + ") complete; removing from awaiting list.");
                            getActiveComputations().remove(id);
                        }
                    }
                } else {
                    logger.debug("Task(id=" + id + ") was not registered! Adding answer to the unregistered answers list.");
                    if (!getUnregisteredMessages().containsKey(id)) {
                        getUnregisteredMessages().put(id, new LinkedList<Message>());
                    }
                    getUnregisteredMessages().get(id).add(message);
                }
            }
            if (c != null) {
                if (c.isComplete()) {
                    logger.debug("Task(id=" + id + ") complete; adding to the completed list.");
                    synchronized (completedLock) {
                        getCompletedComputation().add(c);
                    }
                }
            }
        } else {
            logger.error("Received unsuported answer (not a MapMessage instance): " + message);
        }
    }

    public void info() {
        synchronized (factoryLock) {
            logger.debug("begin FACTORIES");
            for (Entry<String, ComputationFactory> entry : getComputationFactory().entrySet()) {
                logger.debug("\ttype: " + entry.getKey() + " class: " + entry.getValue().getClass());
            }
            logger.debug("end FACTORIES");
        }
        synchronized (activeComputations) {
            logger.debug("begin ACTIVE COMPs");
            for (Entry<String, Computation> entry : getActiveComputations().entrySet()) {
                logger.debug("\tid: " + entry.getKey() + " class: " + entry.getValue().getClass() +
                        " receive: " + entry.getValue().getCurrentQuantity() +
                        " expect: " + entry.getValue().getQuantity());
            }
            logger.debug("end ACTIVE COMPs");
            logger.debug("begin UNREGISTERED MSGs");
            for (Entry<String, List<Message>> entry : getUnregisteredMessages().entrySet()) {
                logger.debug("\tid: " + entry.getKey() + " quantity: " + entry.getValue().size());
            }
            logger.debug("end UNREGISTERED MSGs");
        }
        synchronized (completedLock) {
            logger.debug("begin COMPLETED COMPs");
            for (Computation comp : getCompletedComputation()) {
                logger.debug("\tid: " + comp.getComputationId() + " class: " + comp.getClass());
            }
            logger.debug("end COMPLETED COMPs");
        }
    }

    private List<Computation> getCompletedComputation() {
        return completedComputation;
    }

    public List<Computation> popAllCompletedComputation() {
        List<Computation> result = new LinkedList<Computation>();
        synchronized (completedLock) {
            result.addAll(getCompletedComputation());
            getCompletedComputation().clear();
        }
        return result;
    }
}
